/*
 * Fcalc - a touch-screen calculator for the GP2X-F200
 * written (at a tube) by flynn@vtech.fr 
 * source GPL'ed
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "calc.h"

#define PI 3.141592653590

CTX ctx ;

void enterdigit(int digit) ;
void removedigit(void) ;
void togglesign(void) ;
void convert_acc(void) ;

void acc_init(void)
{
  ctx.s_exp = 0 ;
  ctx.sign = ' ' ;
  memset(ctx.mant, ' ', sizeof(ctx.mant)) ;
  ctx.mant[sizeof(ctx.mant)-2] = '0' ;
  ctx.mant[sizeof(ctx.mant)-1] = 0 ;
  strcpy(ctx.exp, "   ") ;
}

void clr(void)
{
  ctx.page = 0 ;
  ctx.acc = 0L ;
  ctx.ope = 0 ;
  ctx.op1 = 0.0L ;
  ctx.key = 0 ;    // keyin' in progress
  ctx.s_ndx = 0 ;
  acc_init() ;
}

void calc_init(void)
{
  clr() ;
  ctx.mem = 0.0L ;
  ctx.fix = -2 ;
  ctx.deg = 0 ;
}

double oper(void)
{
  double ret ;
  switch(ctx.ope)
  {
   case 0: ret=  ctx.acc ; break ;
   case 1: ret = ctx.op1/ctx.acc ; break ;
   case 2: ret = ctx.op1*ctx.acc ; break ;
   case 3: ret = ctx.op1-ctx.acc ; break ; 
   case 4: ret = ctx.op1+ctx.acc ; break ;
   case 5: ret = pow(ctx.op1, ctx.acc) ; break ;
   case 6: ret = pow(ctx.op1, 1.0/ctx.acc) ; break ;
  }
  ctx.ope = 0 ;
  ctx.acc = ret ;
  return ret ;
}

void pop()
{
  if (ctx.s_ndx<=0) return ;
  ctx.s_ndx-- ;
  ctx.acc = oper() ; 
  ctx.lastx = ctx.acc ; 
  ctx.op1 = ctx.s_op1[ctx.s_ndx] ;
  ctx.ope = ctx.s_op[ctx.s_ndx] ;
}

void push()
{
  if (ctx.s_ndx >= MAXS) return ;
  ctx.s_op1[ctx.s_ndx] = ctx.op1 ;
  ctx.s_op[ctx.s_ndx] = ctx.ope ;
  ctx.ope = 0 ;
  ctx.s_ndx++ ;
}

void convert(void)
{
  char buf[32] ;
  strcpy(buf, ctx.mant) ;
  if (strcmp(ctx.exp, "   "))
  {
    strcat(buf, "e") ; 
    char *p = ctx.exp ;
    while(*p==' ') p++ ;
    strcat(buf, p) ;
  }
  ctx.acc = atof(buf) ;
  if (ctx.sign != ' ') ctx.acc *= -1.0 ;
  printf("buf=%s, mant=%s exp=%s acc=%le op1=%le\n", buf, ctx.mant, ctx.exp, ctx.acc, ctx.op1) ;
}

void convert_acc_1(double acc, char *buf)
{
  sprintf(buf, "%18le", acc) ;
  printf("buf=%s\n", buf) ;
  char *p = strchr(buf, 'e') ;
  if (p!=NULL)
  {
    *p=0 ;
    p++ ;
    strcpy(ctx.exp, p) ;
  }
}

void convert_acc(void)
{
  char buf[132] ;
  double acc = ctx.acc ;
  strcpy(ctx.exp, "   ") ;
  ctx.sign = ' ' ;
  if (ctx.acc < 0)
  {
    togglesign() ;
    acc *= -1.0 ;
  } ; 
  if (ctx.fix>=0)
  {
    char f[10] ;
    sprintf(f, "%%.%dlf", ctx.fix) ;
    sprintf(buf, f, acc) ;
  } else {
    if (ctx.fix==-2)
    {
      sprintf(buf, "%16lf", acc) ;
      int l = strlen(buf)-1 ;
      if (strlen(buf) > sizeof(ctx.mant)-1)
      {
        convert_acc_1(acc, buf) ;
      } else {
        while(l>0 && buf[l]=='0') l-- ;
        if (buf[l]!='.') l++ ;
        buf[l]=0 ;
      }
  printf("_acc2 sign=%c mant=%s exp=%s acc=%le op1=%le\n", 
	ctx.sign, ctx.mant, ctx.exp, ctx.acc, ctx.op1) ;
    } else {
      convert_acc_1(acc, buf) ;
    }  
  }
  strcpy(ctx.mant, buf) ;
  printf("_acc sign=%c mant=%s exp=%s acc=%le op1=%le\n", 
	ctx.sign, ctx.mant, ctx.exp, ctx.acc, ctx.op1) ;
}

void fact(void)
{
  int nacc = (int)trunc(ctx.acc) ;
  if ((nacc>100) || (nacc<=0)) return ;
  ctx.acc=1.0 ;
  for (int i=2; i<=nacc; ++i) ctx.acc *= i ;
}

double ang(double val)
{
  if (ctx.deg==0) return val ;
  return val*2*PI/360.0 ;
}

double aang(double val)
{
  if (ctx.deg==0) return val ;
  return val*180.0/PI ; 
}

void calc(unsigned cde)
{
  int digit = -1 ;
  if (ctx.page==0)
  {
    switch(cde)
    {
      case 11: ctx.acc = ctx.acc*ctx.acc; break ;
      case 12: ctx.acc = sqrt(ctx.acc) ; break ;
      case 13: ctx.op1 = oper(); ctx.ope = 5 ; break ;
      case 14: ctx.op1 = oper(); ctx.ope = 6 ; break ;
      case 15: ctx.page ^= 1 ; break ;

      case 21: ctx.acc = exp(ctx.acc) ; break ;
      case 22: ctx.acc = log(ctx.acc) ; break ;
      case 23: ctx.acc = pow(10, ctx.acc) ; break ;
      case 24: ctx.acc = log10(ctx.acc) ; break ;
      case 25: clr() ; break ;

      case 31: fact() ;break ;
      case 32: ctx.acc = 1.0/ctx.acc ; break ;
      case 33: 
	switch(ctx.ope)
	{
	  case 0: // % monadique ?
		ctx.acc /= 100.0 ;
		break ;
	  case 1: // /x%
		ctx.acc = ctx.op1/ctx.acc/100.0 ;
		break ;
	  case 2: // *x%
		ctx.acc = ctx.op1*ctx.acc/100.0 ;
		break ;
	  case 3: // -x%
		ctx.acc = ctx.op1*(1.0-ctx.acc/100.0) ;
		break ;
	  case 4: // +x% 
		ctx.acc = ctx.op1*(1.0+ctx.acc/100.0) ;
		break ;
	}
	ctx.ope=0 ;
	break ;
      case 34: ctx.acc = PI ; break ;
      case 35: ctx.acc = ctx.mem ; break ;

      case 41: digit = 7 ; break ;// key 7
      case 42: digit = 8 ; break ;// key 8
      case 43: digit = 9 ; break ;// key 9
      case 44: ctx.op1 = oper(); ctx.ope = 1 ; break ;
      case 45: ctx.mem = ctx.acc ; break ;

      case 51: digit = 4 ; break ;// key 4
      case 52: digit = 5 ; break ;// key 5
      case 53: digit = 6 ; break ;// key 6
      case 54: ctx.op1 = oper(); ctx.ope = 2 ; break ;
      case 55: ctx.acc = ctx.lastx ; break ;

      case 61: digit = 1 ; break ;// key 1
      case 62: digit = 2 ; break ;// key 2
      case 63: digit = 3 ; break ;// key 3
      case 64: ctx.op1 = oper(); ctx.ope = 3 ; break ;
      case 65: push() ; break ;

      case 71: digit = 0 ; break ;// key 0
      case 72: digit=10;  break ;
      case 73: togglesign(); return ;
      case 74: ctx.op1 = oper(); ctx.ope = 4 ; break ;
      case 75: pop() ; break ;

      case 81: quit=1 ; break ;// key 0
      case 82: ctx.s_exp=1; return ;
      case 83: removedigit() ; return ;
      // case 84: // does not exist
      case 85: 
	while(ctx.s_ndx>0) pop() ;
	ctx.acc = oper() ; 
	ctx.lastx = ctx.acc ; 
	break ;
    }
  } else {
    switch(cde)
    {
      case 11: ctx.acc = sin(ang(ctx.acc)); break ;
      case 12: ctx.acc = cos(ang(ctx.acc)) ; break ;
      case 13: ctx.acc = tan(ang(ctx.acc)) ; break ;
      case 14: ctx.deg ^= 1 ; break ;
      case 15: ctx.page ^= 1 ; break ;

      case 21: ctx.acc = aang(asin(ctx.acc)) ; break ;
      case 22: ctx.acc = aang(acos(ctx.acc)) ; break ;
      case 23: ctx.acc = aang(atan(ctx.acc)) ; break ;
      case 24: if (++ctx.fix>9) ctx.fix=-2 ; break ;
      case 25: clr() ; break ;

      case 31: break ;
      case 32: break ;
      case 33: break ;
      case 34: ctx.acc = PI ; break ;
      case 35: ctx.acc = ctx.mem ; break ;

      case 41: digit = 7 ; break ;// key 7
      case 42: digit = 8 ; break ;// key 8
      case 43: digit = 9 ; break ;// key 9
      case 44: ctx.op1 = oper(); ctx.ope = 1 ; break ;
      case 45: ctx.mem = ctx.acc ; break ;

      case 51: digit = 4 ; break ;// key 4
      case 52: digit = 5 ; break ;// key 5
      case 53: digit = 6 ; break ;// key 6
      case 54: ctx.op1 = oper(); ctx.ope = 2 ; break ;
      case 55: ctx.acc = ctx.lastx ; break ;

      case 61: digit = 1 ; break ;// key 1
      case 62: digit = 2 ; break ;// key 2
      case 63: digit = 3 ; break ;// key 3
      case 64: ctx.op1 = oper(); ctx.ope = 3 ; break ;
      case 65: push(); break ;

      case 71: digit = 0 ; break ;// key 0
      case 72: digit=10;  break ;
      case 73: togglesign(); return ;
      case 74: ctx.op1 = oper(); ctx.ope = 4 ; break ;
      case 75: pop(); break ;

      case 81: quit=1 ; break ;// key 0
      case 82: ctx.s_exp=1 ; return ;
      case 83: removedigit(); return ;
      // case 84: // does not exist
      case 85: 
	while(ctx.s_ndx>0) pop() ;
	ctx.acc = oper() ; 
	ctx.lastx = ctx.acc ; 
	break ;
    }
  }
  if (digit>=0)
  {
    if (ctx.key==0) acc_init() ;
    enterdigit(digit) ;
    ctx.key=1 ;
  } else {
    convert_acc() ;
    ctx.key=0 ;
  }
}

void togglesign(void)
{
  if (ctx.s_exp==0)
  {
    if (ctx.sign=='-')
    {
      ctx.sign = ' ' ;
    } else {
      ctx.sign = '-' ;
    }
  } else {
    if (ctx.exp[0]=='-')
    {
      ctx.exp[0] = ' ' ;
    } else {
      ctx.exp[0] = '-' ;
    }
  }
  convert() ;
}

void enterdigit(int digit)
{
  int dot=0 ;
  #define L1 (sizeof(ctx.mant)-2)
  if (digit<0) return ;
  if (ctx.s_exp==0)
  {
printf("laaa>%c<\n", ctx.mant[0]) ;
    if (ctx.mant[0] != ' ') return ;
printf("lbbb\n") ;
    dot = (strchr(ctx.mant, '.') != NULL); 
    if (dot && (digit==10)) return ;
    if ((digit==10) || (ctx.mant[L1-1] != ' ') ||
	 (ctx.mant[L1] != '0'))
      memmove(ctx.mant, ctx.mant+1, sizeof(ctx.mant)-1) ;
    if (digit==10)
    {
      ctx.mant[L1] = '.' ; 
    } else
      ctx.mant[L1] = '0'+digit ;
    
  } else {
    if (digit != 10)
    {
      if (ctx.exp[1] == ' ') ctx.exp[1]='0'+digit ; else 
      if (ctx.exp[2] == ' ') ctx.exp[2]='0'+digit ; else 
      {
	  ctx.exp[1] = ctx.exp[2] ;
	  ctx.exp[2] = '0'+digit ;
      }
    }
  }
printf("laaa %s %s\n", ctx.mant, ctx.exp) ;
  convert() ; 
}

void removedigit()
{
  if (ctx.key==0) return ;
  if (ctx.s_exp==0)
  {
    int l = strlen(ctx.mant)-1 ;
    if (l<0) return ;
    ctx.mant[l] = 0 ;
printf("mant l=%d, m=%s\n", l, ctx.mant) ;
    memmove(ctx.mant+1, ctx.mant, l) ;
    if (ctx.mant[l] == ' ') ctx.mant[l] = '0' ;
  } 
  convert() ;
}


